package jp.ac.kyutech.SRP;
// -*- Java -*-
/*!
 * @file  OGSSImpl.java
 * @brief OGSS
 * @date  $Date$
 *
 * $Id$
 */

import java.io.IOException;
import java.net.ConnectException;
import java.net.Socket;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;

import jp.go.aist.rtm.RTC.DataFlowComponentBase;
import jp.go.aist.rtm.RTC.Manager;
import jp.go.aist.rtm.RTC.port.ConnectorBase.ConnectorInfo;
import jp.go.aist.rtm.RTC.port.ConnectorDataListenerT;
import jp.go.aist.rtm.RTC.port.ConnectorDataListenerType;
import jp.go.aist.rtm.RTC.port.InPort;
import jp.go.aist.rtm.RTC.port.OutPort;
import jp.go.aist.rtm.RTC.util.DataRef;
import ogss.AppFrame;

import org.apache.commons.logging.Log;

import RTC.ReturnCode_t;
import RTC.TimedOctetSeq;
import RTC.TimedString;

/*!
 * @class OGSSImpl
 * @brief OGSS
 *
 */
public class OGSSImpl extends DataFlowComponentBase {

    protected static Log log = org.apache.commons.logging.LogFactory.getLog(OGSSImpl.class);
    
    protected static BlockingQueue<String> s_resultQueue;

    static {
        s_resultQueue = null;
    }

    protected static BlockingQueue<String> getResultQueue() {
    	if (s_resultQueue == null) {
    		synchronized(OGSSImpl.class) {
    			if (s_resultQueue == null) {
    				s_resultQueue = new java.util.concurrent.LinkedBlockingQueue<String>();
    			}
    		}
    	}
    	return s_resultQueue;
    }
    
    /**
     * {@link ogss.Juli}特Fʂ󗝂܂B
     * {@link ogss.Juli}singletonp^[ŎĂ邽߁Ã\bhstaticłB
     * @param string
     */
    public static void enqueueRecognitionResult(String string) {
        try {
            synchronized (OGSSImpl.class) {
                if (s_resultQueue != null) {
                    s_resultQueue.put(string);
                    log.debug(String.format("enqeueue result: ", string));
                } else {
                    log.debug(String.format("trash result: ", string));
                }
            }
        } catch (InterruptedException ie) {
            ie.printStackTrace();
        }
    }
    
  /*
     * ! @brief constructor @param manager Maneger Object
     */
    public OGSSImpl(Manager manager) {
        super(manager);
        // <rtc-template block="initializer">
        m_data_val = new TimedOctetSeq();
        m_data = new DataRef<TimedOctetSeq>(m_data_val);
        m_dataIn = new InPort<TimedOctetSeq>("data", m_data);
        m_activategrammar_val = new TimedString();
        m_activategrammar = new DataRef<TimedString>(m_activategrammar_val);
        m_activategrammarIn = new InPort<TimedString>("activategrammar", m_activategrammar);
        m_status_val = new TimedString();
        m_status = new DataRef<TimedString>(m_status_val);
        m_statusOut = new OutPort<TimedString>("status", m_status);
        m_result_val = new TimedString();
        m_result = new DataRef<TimedString>(m_result_val);
        m_resultOut = new OutPort<TimedString>("result", m_result);
        m_log_val = new TimedOctetSeq();
        m_log = new DataRef<TimedOctetSeq>(m_log_val);
        m_logOut = new OutPort<TimedOctetSeq>("log", m_log);
        // </rtc-template>
    }

    /**
     *
     * The initialize action (on CREATED->ALIVE transition)
     * formaer rtc_init_entry() 
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
    @Override
    protected ReturnCode_t onInitialize() {
        // Registration: InPort/OutPort/Service
        // <rtc-template block="registration">
        // Set InPort buffers
        addInPort("data", m_dataIn);
        addInPort("activategrammar", m_activategrammarIn);
        
        m_dataIn.addConnectorDataListener(ConnectorDataListenerType.ON_BUFFER_WRITE, 
                new ConnectorDataListenerT<TimedOctetSeq>(TimedOctetSeq.class) {
            @Override
            public void operator(ConnectorInfo info, TimedOctetSeq data){
            	onData(info, data);
            }
        });
        
        // Set OutPort buffer
        addOutPort("status", m_statusOut);
        addOutPort("result", m_resultOut);
        addOutPort("log", m_logOut);
        // </rtc-template>
        return super.onInitialize();
    }
    
    protected void onData(ConnectorInfo info, TimedOctetSeq data) {
    	// f[^|[gɗoCg񂪓܂
    	if (m_audioSocket != null && m_audioSocket.isConnected()) {
    		try {
    			java.io.OutputStream out = m_audioSocket.getOutputStream();

    			byte[] octets = data.data;
    			out.write(int2octet(octets.length));
    			out.write(octets);
    			
    			out.flush();
    		}
    		catch (IOException e) {
    			log.warn(e.getMessage(), e);
    			try {
    				m_audioSocket.close();
    			} catch (IOException e2) {
        			log.warn(e.getMessage(), e2);
    			}
    			m_audioSocket = null;
    			// Đڑs܂
    	        java.util.concurrent.Executors.newSingleThreadExecutor().submit(new JuliusSocketConnector());
    		}
    	}
    }

    protected byte[] int2octet(int length) {
    	byte[] data = new byte[4];
    	for (int i = 0; i < 4; i++) {
    		data[i] = (byte)(length & 0xFF);
    		length = length >> 8;
    	}
    	return data;
    }

    /***
     *
     * The finalize action (on ALIVE->END transition)
     * formaer rtc_exiting_entry()
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onFinalize() {
//        return super.onFinalize();
//    }

    /***
     *
     * The startup action when ExecutionContext startup
     * former rtc_starting_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
    @Override
    protected ReturnCode_t onStartup(int ec_id) {
        // juli
        AppFrame appFrame = new AppFrame();
        appFrame.setVisible(true);
        java.util.concurrent.Executors.newSingleThreadExecutor().submit(new JuliusSocketConnector());
        return super.onStartup(ec_id);
    }

    /***
     *
     * The shutdown action when ExecutionContext stop
     * former rtc_stopping_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
    @Override
    protected ReturnCode_t onShutdown(int ec_id) {
        closeJuliusAudioSocket();
        closeJuliusModuleSocket();
        return super.onShutdown(ec_id);
    }

    /***
     *
     * The activated action (Active state entry action)
     * former rtc_active_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
    @Override
    protected ReturnCode_t onActivated(int ec_id) {
    	java.util.concurrent.Executors.newSingleThreadExecutor().submit(new Runnable() {
			@Override
			public void run() {
				while(true) {
					try {
						String result = null;
						result = getResultQueue().poll();
						if (result != null) {
							log.debug(String.format("sending result: ", result));
							m_result_val.data = result;
							m_result_val.tm = new RTC.Time(0, 0);
							m_resultOut.write(m_result_val);
						}
						else {
							sleep(500);
						}
					}
					catch (Exception e) {
						log.warn("Fail to send the recognition result.", e);
					}
				}
			}
			
			protected void sleep(long timeInMillis) {
				try {
					Thread.sleep(timeInMillis);
				} catch (InterruptedException e) {
					// 荞݂ꍇ͖܂B
				}
			}
		});
    	
        return super.onActivated(ec_id);
    }

    /***
     *
     * The deactivated action (Active state exit action)
     * former rtc_active_exit()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onDeactivated(int ec_id) {
//        return super.onDeactivated(ec_id);
//    }

    /***
     *
     * The execution action that is invoked periodically
     * former rtc_active_do()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
    @Override
    protected ReturnCode_t onExecute(int ec_id) {
    	return super.onExecute(ec_id);
    };
    
    /***
     *
     * The aborting action when main logic error occurred.
     * former rtc_aborting_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//  @Override
//  public ReturnCode_t onAborting(int ec_id) {
//      return super.onAborting(ec_id);
//  }

    /***
     *
     * The error action in ERROR state
     * former rtc_error_do()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    public ReturnCode_t onError(int ec_id) {
//        return super.onError(ec_id);
//    }

    /***
     *
     * The reset action that is invoked resetting
     * This is same but different the former rtc_init_entry()
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onReset(int ec_id) {
//        return super.onReset(ec_id);
//    }

    /***
     *
     * The state update action that is invoked after onExecute() action
     * no corresponding operation exists in OpenRTm-aist-0.2.0
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onStateUpdate(int ec_id) {
//        return super.onStateUpdate(ec_id);
//    }

    /***
     *
     * The action that is invoked when execution context's rate is changed
     * no corresponding operation exists in OpenRTm-aist-0.2.0
     *
     * @param ec_id target ExecutionContext Id
     *
     * @return RTC::ReturnCode_t
     * 
     * 
     */
//    @Override
//    protected ReturnCode_t onRateChanged(int ec_id) {
//        return super.onRateChanged(ec_id);
//    }
//
    
    protected boolean connectJuliusAudioSocket() {
//        if (true)
//            return true;
        // open audio socket
        try {
            m_audioSocket = new Socket();
            m_audioSocket.connect(new java.net.InetSocketAddress("localhost", 30001));
            log.info("Connected to julius audio socket");
        } catch (ConnectException e) {
        	log.trace("Julius audio socket is closed", e);
        } catch (Exception e) {
            log.error("Fail to connect to julius audio socket", e);
        }
        return (m_audioSocket != null && m_audioSocket.isConnected());
    }

    protected boolean connectJuliusModuleSocket() {
        if (true)
            return true;
        // open module socket
        try {
            m_moduleSocket = new Socket();
            m_moduleSocket.connect(new java.net.InetSocketAddress("localhost", 30002));
            log.info("Connected to julius module socket");
        } catch (ConnectException e) {
            log.trace("Julius module socket is closed", e);
        } catch (Exception e) {
            log.error("Fail to connect to julius module socket", e);
        }
        return (m_moduleSocket != null && m_moduleSocket.isConnected());
    }

    protected void closeJuliusAudioSocket() {
        // close audio socket
        try {
            if (m_audioSocket != null && m_audioSocket.isConnected()) {
                m_audioSocket.close();
            }
        } catch (Exception e) {
            log.error("Fail to close to julius audio socket", e);
        }
    }
    
    protected void closeJuliusModuleSocket() {
        // close module socket
        try {
            if (m_moduleSocket != null && m_moduleSocket.isConnected()) {
                m_moduleSocket.close();
            }
        } catch (Exception e) {
            log.error("Fail to close to julius module socket", e);
        }
    }
    
    // DataInPort declaration
    // <rtc-template block="inport_declare">
    protected TimedOctetSeq m_data_val;
    protected DataRef<TimedOctetSeq> m_data;
    /*!
     */
    protected InPort<TimedOctetSeq> m_dataIn;

    protected TimedString m_activategrammar_val;
    protected DataRef<TimedString> m_activategrammar;
    /*!
     */
    protected InPort<TimedString> m_activategrammarIn;

    
    // </rtc-template>

    // DataOutPort declaration
    // <rtc-template block="outport_declare">
    protected TimedString m_status_val;
    protected DataRef<TimedString> m_status;
    /*!
     */
    protected OutPort<TimedString> m_statusOut;

    protected TimedString m_result_val;
    protected DataRef<TimedString> m_result;
    /*!
     */
    protected OutPort<TimedString> m_resultOut;

    protected TimedOctetSeq m_log_val;
    protected DataRef<TimedOctetSeq> m_log;
    /*!
     */
    protected OutPort<TimedOctetSeq> m_logOut;

    
    // </rtc-template>

    // CORBA Port declaration
    // <rtc-template block="corbaport_declare">
    
    // </rtc-template>

    // Service declaration
    // <rtc-template block="service_declare">
    
    // </rtc-template>

    // Consumer declaration
    // <rtc-template block="consumer_declare">
    
    // </rtc-template>

    protected Socket m_audioSocket;
    
    protected Socket m_moduleSocket;
    
    /**
     * JuliusƉf[^\PbgƃW[f[^\Pbgڑ܂
     */
    protected class JuliusSocketConnector implements Callable<Object> {

        public Object call() throws Exception {
            boolean audioConnected = (m_audioSocket != null && m_audioSocket.isConnected()), moduleConnected = (m_moduleSocket != null && m_moduleSocket.isConnected());
            while (audioConnected == false || moduleConnected == false) {
                if (audioConnected == false) {
                    closeJuliusAudioSocket();
                    audioConnected = connectJuliusAudioSocket();
                }
                if (moduleConnected == false) {
                    closeJuliusModuleSocket();
                    moduleConnected = connectJuliusModuleSocket();
                }
                Thread.sleep(200);
            }
            return null;
        }
    }

    public jp.go.aist.rtm.RTC.port.PortAdmin get_port_admin() {
        return this.m_portAdmin;
    }
}    

